/***************************************************************************
 *   Copyright (C) 2006-2022 by Helge Kortendieck  alan@kasmithee.de       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "qaplay.h"

Qaplay::Qaplay(QSystemTrayIcon *parent) : QSystemTrayIcon(parent)
{
	slClArgs = QApplication::arguments();     // get command line args
	sSongCur = slClArgs.at(1);

	QFileInfo qfi(sSongCur);
	if (qfi.exists() == false or qfi.isReadable() == false) {
		QMessageBox::warning(0, tr("qaplay error - file not readable!"),
		                        tr("The file you specified for playing does either not exist or is not readable!\n"
									"%1\n \n"
									"Please check the name or whether you have read permission for the file.\n \n"
		                            "The program will exit after you clicked 'OK'...").arg(sSongCur));
		exit(0);
	}

	if (QSystemTrayIcon::isSystemTrayAvailable() == false) {
		QMessageBox::warning(0, tr("qaplay error - system tray not available!"),
		                        tr("Ooops, there is no system tray available on this monitor/(virtual)screen combination!\n"
		                            "Until I figure out to use the standard monitor/(virtual)screen to show the system tray icon we will bail out here!\n"
		                            "Sorry - trying harder in the future...\n"
		                            "The program will exit after you clicked 'OK'..."));
		exit(0);
	}

	setIcon(QIcon("/usr/share/qaplay/icons/scalable/qaplay.svg"));
	setToolTip(tr("Right-click and select 'Quit' to stop playing:\n%1").arg(sSongCur));

	actionsDef();                   // define menu actions

	iM3u = 2;
	menu = new QMenu;
	menuCreate();                   // create menu

	bAh = false;
	readDir();
}

Qaplay::~Qaplay()
{
}

void Qaplay::quit()
{
	exit(0);
}

void Qaplay::actionsDef()
{
	actHelp = new QAction(QIcon("/usr/share/qaplay/icons/scalable/help.svg"), tr("&Help"), this);
	actHelp->setStatusTip(tr("Show Help") );
	connect(actHelp, SIGNAL(triggered()), this, SLOT(ahHelp()));

	actAbout = new QAction(QIcon("/usr/share/qaplay/icons/scalable/about.svg"), tr("&About"), this);
	actAbout->setStatusTip(tr("Show about"));
	connect(actAbout, SIGNAL(triggered()), this, SLOT(ahAbout()));

	actPrev = new QAction(QIcon("/usr/share/qaplay/icons/scalable/prev.svg"), tr("&Previous song"), this);
	actPrev->setStatusTip(tr("Previous song"));
	connect(actPrev, SIGNAL(triggered()), this, SLOT(playPrev()));

	actNext = new QAction(QIcon("/usr/share/qaplay/icons/scalable/next.svg"), tr("&Next song"), this);
	actNext->setStatusTip(tr("Next song"));
	connect(actNext, SIGNAL(triggered()), this, SLOT(playNext()));

	actStop = new QAction(QIcon("/usr/share/qaplay/icons/scalable/stop.svg"), tr("&Stop playing"), this);
	actStop->setStatusTip(tr("Stop playing"));
	connect(actStop, SIGNAL(triggered()), this, SLOT(playStop()));

	actQuit = new QAction(QIcon("/usr/share/qaplay/icons/scalable/quit.svg"), tr("&Quit"), this);
	actQuit->setStatusTip(tr("Quit playing"));
	connect(actQuit, SIGNAL(triggered()), this, SLOT(quit()));

	actPlayListShow = new QAction(QIcon("/usr/share/qaplay/icons/scalable/playlist.svg"), tr("Show pla&ylist"), this);
	actPlayListShow->setStatusTip(tr("Show playlist"));
	connect(actPlayListShow, SIGNAL(triggered()), this, SLOT(playListShow()));

	actPlayListHide = new QAction(QIcon("/usr/share/qaplay/icons/scalable/playlist.svg"), tr("Hide play&list"), this);
	actPlayListHide->setStatusTip(tr("Hide playlist"));
	connect(actPlayListHide, SIGNAL(triggered()), this, SLOT(playListHide()));

	actInfo = new QAction(QIcon("/usr/share/qaplay/icons/scalable/info.svg"), tr("File/Song I&nfo"), this);
	actInfo->setStatusTip(tr("Show file/song info"));
	connect(actInfo, SIGNAL(triggered()), this, SLOT(infoShow()));

	actFavAdd = new QAction(QIcon("/usr/share/qaplay/icons/scalable/favAdd.svg"), tr("A&dd to Favorites"), this);
	actFavAdd->setStatusTip(tr("Add to favorites list"));
	connect(actFavAdd, SIGNAL(triggered()), this, SLOT(favAdd()));

	actFavPlay = new QAction(QIcon("/usr/share/qaplay/icons/scalable/favPlay.svg"), tr("Pla&y Favorites"), this);
	actFavPlay->setStatusTip(tr("Play favorites list"));
	connect(actFavPlay, SIGNAL(triggered()), this, SLOT(favPlay()));
}

void Qaplay::menuCreate()
{
	menu->clear();
	menu->addAction(QIcon("/usr/share/qaplay/icons/qaplay.png"), tr("qaplay"));
	menu->addSeparator();
	
	menu->addAction(actHelp);				// show help widget
	menu->addAction(actAbout);				// show about widget
	menu->addSeparator();
	
	menu->addAction(actPrev);				// play previous song
	menu->addAction(actNext);				// play next song
	menu->addAction(actStop);				// stop playing
	menu->addSeparator();
	
	menu->addAction(actInfo);				// show file/song info
	if (iM3u > 0) {
		menu->addSeparator();
		menu->addAction(actFavAdd);				// add song to favorites list
		if (m3uExists() == 1) {
			menu->addAction(actFavPlay);			// play favorites list
		}
	}
	menu->addSeparator();
	
	menu->addAction(actQuit);				// quit
	setContextMenu(menu);
}

void Qaplay::toolTipUpd()					// update tooltip with song currently played
{
	setToolTip(tr("Right-click and select 'Quit' to stop playing:\n%1").arg(slSongs.at(iSongCurListIdx)));
}

void Qaplay::readDir()						// get all songs in dir, calling path can be file, dir or *.m3u
{
	QFileInfo fi(sSongCur);
	if (fi.isDir() == true) {					// called with directory
		sD = sSongCur;
		iInvokedBy = 1;
	} else if (fi.suffix().toLower() == "m3u") {	// called with m3u list
		iInvokedBy = 3;
	} else {									// called with file/song
		sD = fi.path();
		iInvokedBy = 2;
	}
	slSongs.clear();

	if (iInvokedBy < 3) {						// called with dir or song, fetch all songs in dir...
		if (sD.right(1) != "/") {
			sD += "/";
		}
	
		QDir dir( sD );								// song directory, look for songs to play in this dir
			slFilters << "*.mp3" << "*.ogg" << "*.flac" << "*.wav" << "*.wma" << "*.aac" << "*.ts" << "*.mkv" << "*.mp4" << "*.webm";	// more?
			dir.setNameFilters(slFilters);
			dir.setFilter(QDir::Files | QDir::Readable);
			dir.setSorting(QDir::Name | QDir::IgnoreCase);
		slD = dir.entryList();

		if (slD.count() > 0) {						// found at least 1 song with filter settings
			for (int i = 0; i < slD.count(); ++i) {
				slSongs.append(sD + slD.at(i));
//cout << "readDir() - add songs: " << slSongs.last().toStdString() << endl;                                               // *****  debug  *****
			}
			bD = false;
			if (iInvokedBy == 2) {
				for (int i = 0; i < slSongs.count(); ++i) {
					if (slSongs.at(i) == sSongCur) {	// song qaplay was called with is among the list of files returned by dir serach
						bD = true;
						iSongCurListIdx = i;
						break;
					}
				}
				if (bD == false) {						// did not found song qaplay was called with in song list
					slSongs.append(sSongCur);			// qaplay was called with different song type we support (and set filters ), let's try anyway, add at end
					iSongCurListIdx = slSongs.count() - 1;
				}
			} else {
				iSongCurListIdx = 0;				// play dir: start with 1st song in dir
			}
		} else {
			slSongs.append(sSongCur);				// qaplay was called with different song type we support (and set filters ), let's try anyway (can now be a dir!)
			iSongCurListIdx = 0;					// qaplay was invoked with this only file in list
		}
	} else {									// qaplay was called with a m3u playlist
		slSongs = m3uRead(sSongCur);
		iSongCurListIdx = 0;						// set index to 1st song
	}
	
	playCreate();
}

//  *****  media player  *****
void Qaplay::playCreate()
{
	player = new QMediaPlayer;
	connect(player, SIGNAL(mediaStatusChanged(QMediaPlayer::MediaStatus)), this, SLOT(playStatusChanged(QMediaPlayer::MediaStatus)));
	playPlay();
}

void Qaplay::playPlay()
{
	player->setMedia(QUrl::fromLocalFile(slSongs.at(iSongCurListIdx)));
//	player->setVolume(50);
	toolTipUpd();
	player->play();
}

void Qaplay::playStatusChanged(QMediaPlayer::MediaStatus state)
{
	switch( state) {
		case QMediaPlayer::EndOfMedia:
		    if (slSongs.count() - 1 > iSongCurListIdx) {
				iSongCurListIdx += 1;
				playPlay();
			} else {							// quit at end of list
				quit();
			}
			break;
		case QMediaPlayer::InvalidMedia:
			showMessage(tr("Invalid Media File!"), tr("Cannot play the file %1").arg(slSongs.at(iSongCurListIdx)),QSystemTrayIcon::Information,3000);
			quit();
			break;
	}
}

void Qaplay::playPrev()
{
	if (iSongCurListIdx > 0) {
		iSongCurListIdx -= 1;
	} else {
		iSongCurListIdx = slSongs.count() - 1;	// roll over, reached begin, start at end of list, last song again...
	}
	player->stop();
	playPlay();
}

void Qaplay::playNext()
{
	if (slSongs.count() - 1 > iSongCurListIdx) {
		iSongCurListIdx += 1;
	} else {
		iSongCurListIdx = 0;					// roll over, reached end, start at 0, first song again...
	}
	player->stop();
	playPlay();
}

void Qaplay::playStop()
{
	player->stop();
	quit();										// for now we stop, as we cannot re-start without playlist !!!!!!!!!!!!!!!!!!!!!!
}

//  *****  show about/help dialog  *****
void Qaplay::ahAbout()						// show about dialog (1)
{
	if (bAh == false) {
		bAh = true;
		ahShow(1);
	}
}

void Qaplay::ahHelp()						// show help dialog (2)
{
	if (bAh == false) {
		bAh = true;
		ahShow(2);
	}
}

void Qaplay::ahShow(int iD)					// show abotu/help dialog
{
	ahw = new aboutHelp();
	connect(this, SIGNAL(sigAh(int)), ahw, SLOT(showAh(int)));
	connect(ahw, SIGNAL(sigDelete()), this, SLOT(ahDelete()));
	emit sigAh(iD);
}

void Qaplay::ahDelete()						// delete about/help dialog
{
	ahw->deleteLater();
	bAh = false;
}

//  *****  show file/song info  *****
void Qaplay::infoShow()						// show song info
{
	siD.sArtist = player->metaData(QMediaMetaData::AlbumArtist).toString();
	siD.sTitle = player->metaData(QMediaMetaData::Title).toString();
	siD.sAlbum = player->metaData(QMediaMetaData::AlbumTitle).toString();
	siD.sYear = player->metaData(QMediaMetaData::Year).toString();
	siD.sComment = player->metaData(QMediaMetaData::Comment).toString();
	siD.sTrack = player->metaData(QMediaMetaData::TrackNumber).toString();
	siD.sGenre = player->metaData(QMediaMetaData::Genre).toString();
	iD = player->metaData(QMediaMetaData::AudioBitRate).toInt() / 1000;
	siD.sBitRate = QString::number(iD);
	siD.sChannels = player->metaData(QMediaMetaData::ChannelCount).toString();
	siD.sCodec = player->metaData(QMediaMetaData::AudioCodec).toString();
	sD = "";
	if (siD.sArtist != "") {
		sD = siD.sArtist;
	}
	if (siD.sTitle != "") {
		sD += " - " + siD.sTitle;
	}
	if (siD.sAlbum != "") {
		sD += " | " + siD.sAlbum;
	}
	if (siD.sTrack != "") {
		sD += " - " + siD.sTrack;
	}
	if (siD.sYear != "") {
		sD += " - " + siD.sYear;
	}
	if (siD.sGenre != "") {
		sD += " | " + siD.sGenre;
	}
	if (siD.sComment != "") {
		sD += " | " + siD.sComment;
	}
	if (siD.sBitRate != "") {
		sD += " | " + siD.sBitRate + " kbps | ";
	}
	sD += infoFormatTime(player->position()) + "/";
	sD += infoFormatTime(player->duration());
	if (siD.sChannels != "") {
		sD += " | " + siD.sChannels;
	}
	if (siD.sCodec != "") {
		if (siD.sCodec.contains("MP3",Qt::CaseInsensitive)) {
			siD.sCodec = "mp3";
		} else if(siD.sCodec.contains("OGG"),Qt::CaseInsensitive) {
			siD.sCodec = "ogg";
		} else if(siD.sCodec.contains("FLAC"),Qt::CaseInsensitive) {
			siD.sCodec = "flac";
		}
		sD += " | " + siD.sCodec;
	}
//cout << "showInfo() : " << sD.toStdString() << endl;                                               // *****  debug  *****
	if (sD != "") {
		showMessage(tr("Song Info"),sD,QSystemTrayIcon::Information,8000);
	} else {
		showMessage(tr("Song Info"),tr("Ooops! Sorry, could not get any song information from the file currently playing."),QSystemTrayIcon::Information,8000);
	}
}

QString Qaplay::infoFormatTime(int iD2)		// format song time - position n duration, expects milliseconds
{
	iD2 = iD2 / 1000;							// turn milliseconds into seconds
	if (iD2 / 60 < 10) {						// get minutes
		sD2 = "0" + QString::number(iD2 / 60) + ":";
	} else {
		sD2 = QString::number(iD2 / 60) + ":";
	}
	iD2 = iD2 - (iD2 / 60 * 60);
	if (iD2 < 10) {						// get minutes
		sD2 += "0" + QString::number(iD2);
	} else {
		sD2 += QString::number(iD2);
	}
	return sD2;
}

//  *****  favorites  *****
void Qaplay::favAdd()						// add song to favorites list
{
	if (iM3u == 0) {							// not really needed ?!
		return;
	}
	iD = m3uExists();
//cout << "favAdd() iD: " << QString::number(iD).toStdString() << endl;                                               // *****  debug  *****
	if (iD == 2) {								// ~/.config/qaplay/qaplay.m3u exists but is not writable
		showMessage(tr("qaplay error"),tr("Sorry, cannot add a song: ~/.config/qaplay/qaplay.m3u is not writable!"),QSystemTrayIcon::Information,8000);
		iM3u = 0;
		menuCreate();
		return;
	}
	if (iD == 0) {								// ~/.config/qaplay/qaplay.m3u does not exists try to create it, can be the first time this is called
		if (m3uCreate() == 0) {					// create new, empty ~/.config/qaplay/qaplay.m3u - abort if not possible
			showMessage(tr("qaplay error"),tr("Sorry, an error occurred while trying to create favorites song list: ~/.config/qaplay/qaplay.m3u !"),QSystemTrayIcon::Information,8000);
			iM3u = 0;
			menuCreate();
			return;
		}
	}
	if (iM3u == 2) {							// try reading ~/.config/qaplay/qaplay.m3u only once
		slM3u = m3uRead(QDir::home().absolutePath() + "/.config/qaplay/qaplay.m3u");
		if (slM3u.size() == 0) {				// is empty, not even the header exists - disable favorites
			showMessage(tr("qaplay error"),tr("Sorry, an error occurred while trying to read favorites song list: ~/.config/qaplay/qaplay.m3u !"),QSystemTrayIcon::Information,8000);
			iM3u = 0;
			menuCreate();
			return;
		} else {
			iM3u = 1;							// let's think we have a valid m3u... - no checks for a valid m3u so far
		}
	}
	if (slM3u.contains(slSongs.at(iSongCurListIdx), Qt::CaseInsensitive) == false) {	// check whether the titel is already in the m3u - do not add it twice
		// add song to m3u
		slM3u.append("#EXTINF:" + QString::number(player->duration()/1000) + "," + player->metaData(QMediaMetaData::AlbumArtist).toString() + " - " + player->metaData(QMediaMetaData::Title).toString());
		slM3u.append(slSongs.at(iSongCurListIdx));
//cout << "favAdd() slM3u.at(1): " << slM3u.at(1).toStdString() << endl;                                               // *****  debug  *****
		if (m3uWrite(slM3u) == 0) {					// error while writing m3u immediately
			showMessage(tr("qaplay error"),tr("Sorry, an error occurred while trying to write favorites song list: ~/.config/qaplay/qaplay.m3u !"),QSystemTrayIcon::Information,8000);
			iM3u = 0;
			menuCreate();
			return;
		}
		showMessage(tr("qaplay info"),tr("Current song added to the favourites playlist!"),QSystemTrayIcon::Information,8000);	// yeah, we did it!
	} else {
		showMessage(tr("qaplay info"),tr("This song is already in the favorites playlist - not adding it again!"),QSystemTrayIcon::Information,8000);	
	}
}

void Qaplay::favPlay()						// play favorites list
{
	player->stop();
	sSongCur = QDir::home().absolutePath() + "/.config/qaplay/qaplay.m3u";
	readDir();	
}

//  *****  song list - show/delete  *****
void Qaplay::playListShow()					// show playlist widget
{

}

void Qaplay::playListHide()					// hide playlist widget
{

}

void Qaplay::playListDelete()				// delete playlist widget
{

}

//#include "qaplay.moc"
